from .log_handler import set_global_logs, clear_global_log
from .settings import logger_settings
from .decorator import FunctionLogger

name = 'django_logger'

LOGGING_FORMAT = logger_settings.LOGGING_FORMAT

__all__ = [LOGGING_FORMAT, FunctionLogger, set_global_logs, clear_global_log]

from setuptools import setup


"""
This module is largely inspired by django-rest-framework settings.

Settings for the django logging are all namespaced in the DJANGO_LOGGER setting.
For example your project's `settings.py` file might look like this:

DJANGO_LOGGER = {
    "IGNORE_PARAMS":
        ['password','username',],
    "APP":
        "APP_NAME",
}

This module provides the `DjangoLoggerSettings` object, that is used to access
Django Logger settings, checking for user settings first, then falling
back to the defaults.
"""
from __future__ import unicode_literals

import copy
import json
import logging

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

from .logformat import BASE_LOGGING_FORMAT

USER_SETTINGS = getattr(settings, "DJANGO_LOGGER", None)

# List of settings that cannot be empty
MANDATORY = (
    "APP_NAME",  # application name
    "LOGGER_NAME",  # logger name
    "MAIL_LOGGER_NAME",  # mail logger name
)

DEFAULTS = {
    "SECURE_PARAMS": (),
}


class DjangoLoggerSettings(object):
    """
    A settings object, that allows Django Logger settings to be accessed as properties.

    """

    def __init__(self, user_settings=None, defaults=None, mandatory=None):

        # Initialises logger properties, as later one's depends on
        # previous one's so don't change the order.

        self.user_settings = user_settings or {}
        self.defaults = defaults or {}
        self.mandatory = mandatory or ()
        self.validate_setting()
        self.user_logger_formatters = {}
        self.HANDLER = self.get_handler()
        self.LOGGING_FORMAT = self.get_logger_format(BASE_LOGGING_FORMAT)
        self.GLOBAL = {}

    def __getattr__(self, attr):

        try:
            # Check if present in user settings
            val = self.user_settings[attr]
        except KeyError:
            # Fall back to defaults

            if attr not in self.defaults.keys():
                raise ImproperlyConfigured("Invalid DjangoLogger setting: %r." % (attr))

            val = self.defaults[attr]

        # Cache the result
        setattr(self, attr, val)

        return val

    def _construct_logger_format(self, format):

        for key, val in format.items():

            if val == {}:
                format[key] = '%(' + key + ')s'
                self.user_logger_formatters[key] = dict

            elif not val:
                format[key] = '%(' + key + ')s'
                self.user_logger_formatters[key] = str

            elif isinstance(val, dict):
                format[key] = self._construct_logger_format(val)

            elif val in (int,):
                format[key] = '%(' + key + ')d'
                self.user_logger_formatters[key] = int

            elif val in (str, dict):
                format[key] = '%(' + key + ')s'
                self.user_logger_formatters[key] = str

            elif val in (float,):
                format[key] = '%(' + key + ').2f'
                self.user_logger_formatters[key] = float

        return format

    def get_logger_format(self, base_format):

        APP_NAME = self.user_settings.get('APP_NAME')
        DATA_KEY = self.user_settings.get('NAMESPACE_KEY', APP_NAME.upper() + "_DATA")

        logging_format = json.dumps(copy.deepcopy(base_format))

        # replace path names as per settings
        logging_format = logging_format \
            .replace('$PATH_APP_NAME', APP_NAME) \
            .replace('$PATH_LOGGER_NAME', self.user_settings.get('LOGGER_NAME')) \
            .replace('$PATH_MAIL_LOGGER_NAME', self.user_settings.get('MAIL_LOGGER_NAME'))

        logging_format = json.loads(logging_format)

        if self.user_settings.get('FILENAME'):
            logging_format['handlers']['console']['filename'] = self.user_settings.get('FILENAME')
            logging_format['handlers']['console']['mode'] = self.user_settings.get('FILEMODE', "w")

        _user_format = copy.deepcopy(self.user_settings.get('LOGGING_FORMAT', {}))
        user_format = {}

        if not _user_format:
            user_format[DATA_KEY] = {"MESSAGE": dict}

        elif "MESSAGE" not in _user_format:
            _user_format.update({"MESSAGE": dict})
            user_format[DATA_KEY] = copy.deepcopy(_user_format)

        else:
            user_format[DATA_KEY] = _user_format

        user_format = self._construct_logger_format(user_format)

        _detailed_log_format = logging_format['formatters']['detailed_log_track']['format']
        _json_detailed_log_format = json.loads(_detailed_log_format)

        _json_detailed_log_format.update(user_format)

        # replace capital or small `message` formatters with correct one
        _detailed_log_format = json.dumps(_json_detailed_log_format) \
            .replace('\n', '').replace(' ', '') \
            .replace('\"%(MESSAGE)s\"', "%(message)s") \
            .replace('\"%(message)s\"', "%(message)s") \
            .replace('\"%(trace)s\"', "%(trace)s")

        # Formatting by finding regex and replacing
        import re
        _detailed_log_format = re.sub('\"%\((?P<name>[a-zA-Z]*)\)(?P<formatter>[d]|.2f)\"',
                                      '%(\g<name>)\g<formatter>',
                                      _detailed_log_format)

        logging_format['formatters']['detailed_log_track']['format'] = _detailed_log_format

        return logging_format

    def get_handler(self):
        # Sets handler for log format if filename is given
        # in configuration FileHandler will be used else
        # StreamHandler

        if self.user_settings.get('FILENAME'):
            return logging.FileHandler
        else:
            return logging.StreamHandler

    def validate_setting(self):

        # validate mandatory settings
        for attr in self.mandatory:
            if not attr in self.user_settings or not self.user_settings.get(attr):
                raise ImproperlyConfigured("DjangoLogger setting: %r is mandatory" % attr)
            # ImproperlyConfigured

        # validate logger format
        if not isinstance(self.user_settings.get('LOGGING_FORMAT', {}), dict):
            raise ImproperlyConfigured("DjangoLogger settings LOGGING_FORMAT must be a %r" % type({}))


logger_settings = DjangoLoggerSettings(USER_SETTINGS, DEFAULTS, MANDATORY)



BASE_LOGGING_FORMAT = {
    'version': 1,
    'disable_existing_loggers': False,

    'filters': {
        'require_debug_false': {
            '()': 'django.utils.log.RequireDebugFalse',
        },
        'require_debug_true': {
            '()': 'django.utils.log.RequireDebugTrue',
        },
    },

    'formatters': {

        'simple': {
            'format': '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
        },

        'detailed_log_track': {
            'format': '''
                        {
                            "APP": "$PATH_APP_NAME",
                            "TIME": "%(time)s",
                            "LEVEL": "%(levelname)s",
                            "LOGGER": "%(name)s",
                            "PATH": "%(pathname)s",
                            "FUNCTION": "%(funcName)s",
                            "LINE": "%(lineno)d",
                            "TRACE": "%(trace)s"
                        }''',  # .replace('$PATH_APP',logger_settings.APP),

            # 'datefmt': '%Y-%m-%d %H:%M:%S',

        },
    },
    'handlers': {

        'console': {
            'level': 'INFO',
            'class': 'django_logger.log_handler.ContextHandler',
            'formatter': 'detailed_log_track'

        },
        'mail_admins': {
            'level': 'INFO',
            # we don't want to send you mails while in dev mode
            'filters': ['require_debug_false'],
            'class': 'django.utils.log.AdminEmailHandler',
            'formatter': 'detailed_log_track',

        },

    },

    'loggers': {
        "$PATH_LOGGER_NAME": {
            'handlers': ['console', ],
            'propagate': False,
            'level': 'INFO',
        },
        # Mail logger should log message to console as well.
        "$PATH_MAIL_LOGGER_NAME": {
            'handlers': ['console', 'mail_admins', ],
            'propagate': False,
            'level': 'INFO',
        },
    },
}


from __future__ import absolute_import

import threading

from django_logger.settings import logger_settings

"""Data-wrangling for loggers"""
import ast
import json
import logging
import traceback

import arrow
import pytz


def clear_global_log():
    """
    Clears global log, this needs to be called before setting global log
    in order to clear if any previous process would have set any logs.

    """
    logger_settings.GLOBAL = dict()


# Helper function for setting value in global log via crequest in case of django
def set_global_logs(values):
    """
    Sets key, value pair from values dict, in global log

    Parameters
    ----------
    values: <dict>
        Dict containing key:value pair, which needs to be set,
        must be present in DJANGO_LOGGER[`LOGGING_FORMAT`] settings

    Returns
    -------

    Exception: If `key` not present in DJANGO_LOGGER[`LOGGING_FORMAT`] or
               `request` is empty

    True: If successfully sets `val` to `key` in values in global log

    """
    for key in values.keys():
        if key not in logger_settings.user_logger_formatters.keys():
            raise Exception('Key {} not present in config defined {}.'.format(key, logger_settings.user_logger_formatters))

    # setting value in global against current thread
    current_thread = threading.current_thread()
    if not logger_settings.GLOBAL.get(current_thread):
        logger_settings.GLOBAL[current_thread] = {}

    for key,val in values.items():
        logger_settings.GLOBAL[current_thread][key] = val

    return True


class ContextualFilter(logging.Filter):
    """
        Overrides loggging.Filter
        to the logging Context. DOES NOT FILTER ANY LOGS, all logs are passed.
    """

    def filter(self, log_record):
        """Filter messages to be logged.

        Parameters
        ----------
        `log_record` : <logging.LogRecord>
            The log-record to be processed and filtered.

        Returns
        -------
        `True`
            if the `log_record` is to be logged.
        `False`
            if the `log_record` is to be dropped.

        Notes
        -----
        Currently we aren't dropping any logs. This function simply transforms
        the log-record by injecting additional data into it, and prepares keys
        for the log-formatter.
        """

        def _get_key_value(key=None, val_type=None):

            value = None
            current_thread = threading.current_thread()

            if logger_settings.GLOBAL.get(current_thread):
                value = logger_settings.GLOBAL[current_thread].get(key)

            if not value:
                if val_type in (int,):
                    value = 0
                elif val_type in (float,):
                    value = 0.0
                elif val_type in (dict,):
                    value = {}
                else:
                    value = ""

            return value

        def _get_trace():
            trace = ""
            if log_record.levelname in ['ERROR', 'CRITICAL']:
                trace = traceback.format_exc().strip()  # Get the recent stack-trace

            return json.dumps(trace)

        def _get_msg():
            try:
                # `loads` and `dumps` for printing JSON structure in a single line
                json_data = json.loads(log_record.msg)
                msg = json.dumps(json_data)
            except ValueError:
                msg = json.dumps({
                    "TEXT": log_record.msg,
                })
            except TypeError:
                # We tried to `json.loads` something that's not a string
                # Let's print its string-representation
                msg = json.dumps({
                    "OBJ": repr(log_record.msg),
                })
            return msg

        def _get_log_time():
            tz_IST = pytz.timezone(pytz.country_timezones['IN'][0])
            return str(arrow.now(tz=tz_IST).format('YYYY-MM-DDTHH:mm:ss.SSSZZ'))

        log_record.trace = _get_trace()
        log_record.msg = _get_msg()
        log_record.time = _get_log_time()

        # setting values for key log_record
        for key, val_type in logger_settings.user_logger_formatters.items():
            setattr(log_record, key, _get_key_value(key, val_type))

        # Clean message
        try:
            # If case of `exception` level is there we don't want trace print it again
            if log_record.exc_info:
                log_record.exc_info = None  # its checked in python logging (__init__.py) #571 in logging for python 3.5

            log_msg = json.loads(log_record.msg)
            msg = log_msg.get('OBJ', None)
            if msg:
                msg_dict = ast.literal_eval(msg)

                _args = msg_dict.get('ARGS', "")

                if not isinstance(_args, str):
                    _args = repr(_args)

                _kwargs = msg_dict.get('KWARGS', "")
                if not isinstance(_kwargs, str):
                    _kwargs = repr(_kwargs)

                if msg_dict:
                    if 'MESSAGE' in msg_dict:
                        log_record.msg = json.dumps({
                            'TEXT': msg_dict.get('MESSAGE', ''),
                            'ARGS': _args,
                            'KWARGS': _kwargs
                        })
                        log_record.pathname = msg_dict.get('_PATH', log_record.pathname)
                        log_record.funcName = msg_dict.get('_FUNCNAME', log_record.funcName)
                        log_record.lineno = msg_dict.get('_LNO', log_record.lineno)
                    else:
                        log_record.msg = json.dumps({})
        except Exception as e:
            log_record.msg = json.dumps({
                "ERROR": "Exception occurred while parsing `{}`".format(e),
                "MESSAGE": log_record.msg
            })
            pass

        return True  # Don't drop any logs


class ContextHandler(logger_settings.HANDLER):
    """
        Overrides logging.StreamHandler

    """

    def __init__(self, *args, **kwargs):
        logger_settings.HANDLER.__init__(self, *args, **kwargs)
        self.addFilter(ContextualFilter())



import copy
import logging

from collections import namedtuple
from .settings import logger_settings

try:
    from inspect import getfullargspec
except ImportError:
    # for python 2.7
    from inspect import getargspec as getfullargspec

logger = logging.getLogger(logger_settings.LOGGER_NAME)


def convert_to_type(type_var, var):
    return type_var(var)


class FunctionLogger(object):
    """
    Function logger to log at entry-exit of a function with arguments and output

    Usage
    -----
    @FunctionLogger(secure_args=['a','b',]
    def sum(a,b):
        return a+b

    When called as sum(1,2) it will be logged as:

    "APP":"FUNDSFLOW_IMPS","TIME":"2018-01-05T15:31:41.961+05:30","LEVEL":"INFO","LOGGER":"imps",
    "PATH":"<ipython-input-2-28c5fd5efdaf>","FUNCTION":"sum","LINE":"1",
    "IMPS_DATA":{.....},"MESSAGE":{"KWARGS": "{}", "ARGS": "NamedArgs(a='*', b='*')", "TEXT": "Entering sum"}},
    "TRACE":""}

    While initialising we can give arguments which has to be masked in secure_args list

    note:: Number of `*` has been limited to max 10, in case of exceeding length of parameter
        it will be truncated to length % 10 number of `*`
    """

    def __init__(self, secure_args=[]):
        self.secure_list = secure_args

        # read from settings if defined there
        try:
            self.secure_list += list(logger_settings.SECURE_PARAMS)
        except AttributeError:
            pass

    def _get_secure_log(self, params, arg_names=None):

        is_kwargs = False

        if type(params) == dict:
            is_kwargs = True
            _params = copy.deepcopy(params)
            secure_params = {}
        else:
            secure_params = []

        if params:
            for arg_name, var in zip(arg_names, params) if arg_names else params.items():

                arg_val = _params[arg_name] if is_kwargs else var

                if isinstance(arg_val, (list, tuple)) and arg_name in self.secure_list:

                    _ = ['*' * (len(str(i)) % 10) for i in arg_val]
                    arg_val = convert_to_type(type(arg_val), _)

                elif isinstance(arg_val, (dict,)):

                    if arg_name in self.secure_list:
                        arg_val = {k: '*' * (len(str(v)) % 10) for k, v in arg_val.items()}
                    else:
                        arg_val = {k: '*' * (len(str(v)) % 10) if k in self.secure_list else v for k, v in
                                   arg_val.items()}

                elif arg_name in self.secure_list:

                    arg_val = '*' * (len(str(arg_val)) % 10)

                if is_kwargs:
                    secure_params[arg_name] = arg_val if isinstance(arg_val, (str, dict, float, int)) else str(arg_val)
                else:
                    secure_params.append(arg_val)

        if is_kwargs:
            # Lets sort dictionary keys so that it prints alphabetically
            return secure_params
        else:
            Args = namedtuple('NamedArgs', arg_names[:len(secure_params)])
            named_secure_params = Args(*secure_params)
            return repr(named_secure_params)

    def __call__(self, fn):

        def wrapped_f(*args, **kwargs):
            log_args = args
            log_kwargs = kwargs

            arg_spec = getfullargspec(fn)

            # take only args from
            # FullArgSpec(args, varargs, varkw, defaults, kwonlyargs, kwonlydefaults, annotations)
            arg_names = arg_spec[0]
            secure_log_args = self._get_secure_log(log_args, arg_names)
            secure_log_kwargs = self._get_secure_log(log_kwargs) if log_kwargs else ""

            logger.info({
                'MESSAGE': 'Entering {}'.format(fn.__name__),
                'ARGS': secure_log_args,
                'KWARGS': secure_log_kwargs,
                '_LNO': fn.__code__.co_firstlineno,
                '_FUNCNAME': fn.__name__,
                '_PATH': fn.__code__.co_filename
            })

            fn_output = fn(*args, **kwargs)

            secured_fn_output = copy.deepcopy(fn_output) if isinstance(fn_output, dict) else fn_output
            # Mask output only if its dict
            if isinstance(fn_output, dict):
                secured_fn_output = {
                    k: '*' * (len(str(v)) % 10) if k in self.secure_list else v for k, v in fn_output.items()
                }
            elif not isinstance(secured_fn_output, str):
                secured_fn_output = repr(secured_fn_output)

            logger.info({
                'MESSAGE': 'Exiting {}'.format(fn.__name__),
                'ARGS': secured_fn_output if secured_fn_output else "",
                '_LNO': fn.__code__.co_firstlineno,
                '_FUNCNAME': fn.__name__,
                '_PATH': fn.__code__.co_filename
            })
            return fn_output

        return wrapped_f



from setuptools import setup, find_packages


with open("README.md", "r") as fh:
    long_description = fh.read()


setup(name='easy-django-logger',
      version='0.1.1',  # version.feature.patch (rc)release-candidate 1
      url='https://github.com/NiranjanSingh/django-logger',
      license='MIT',
      author='Niranjan Singh',
      author_email='niranjan32331@gmail.com',
      description='Django logger to log formatted logs parsable via elastic search and visualize via kibana.',
      packages=find_packages('.', exclude='logger_test'),
      install_requires=[
          'Django>=1.8,<2',
          'arrow>=0.12.1',
          'pytz>=2017.3',
      ],
      long_description=long_description,
      long_description_content_type="text/markdown",
      zip_safe=False)  # As there is only one package `django_logger` zipping it won't effect much.
